# 
# -*- coding: us-ascii -*-
# generated by wxGlade 0.6.5 on Sun Feb 12 16:07:33 2012
# version 1
# this version can also measure distances

from PyQt5 import QtWidgets, QtCore, QtGui
import sys
#import wx
import os

import LT.box as B

#import pdb

# begin wxGlade: extracode
# end wxGlade

import matplotlib.pyplot as plt
#import matplotlib.patches as mpatches

from matplotlib.figure import Figure
import numpy as np

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5 import NavigationToolbar2QT

#----------------------------------------------------------------------
# helper functions
#----------------------------------------------------------------------

def angle(sint, cost):
    # calculate the angle 0 < theta < 2 pi if sin and cos are given
    ang = 0.
    # check data boundaries
    if cost > 1.: 
        cost = 1.
    if cost < -1. : 
        cost = -1.
    if sint > 1.:
        sint = 1.
    if sint < -1.:
        sint= - 1.
    arg1=np.arccos(cost)
# 
# find the quadrant
# 
    q1 = (sint >= 0.) and (cost >= 0.)
    q2 = (sint >= 0.) and (cost < 0.)
    q3 = (sint < 0.) and (cost < 0.)
    q4 = (sint < 0.) and (cost >= 0.)
    if (q1 or q2):
        ang=arg1
    if (q3 or q4):
        ang=2.*np.pi-arg1 
    return ang



class NumberDialog(QtWidgets.QDialog):
     def __init__(self, data, parent = None, title = 'Enter parameters', labels=['First Label','Second Label'],
                  keys=['first','second'], about_txt = 'about_txt'):
          
          self.parent=parent
          self.keys=keys
          self.data=data
          QtWidgets.QDialog.__init__(self, parent)
          self.setWindowFlags(self.windowFlags() ^ QtCore.Qt.WindowContextHelpButtonHint)
          self.setWindowTitle(title)
          self.layout=QtWidgets.QGridLayout(self)
          about = QtWidgets.QLabel(about_txt, self)
          ok=QtWidgets.QPushButton('Ok', self)
          ok.setDefault(True)
          cancel=QtWidgets.QPushButton('Cancel', self)
          ok.clicked.connect(self.OnOk)
          cancel.clicked.connect(self.OnCancel)
          sepline=QtWidgets.QFrame()
          sepline.setFrameShape(QtWidgets.QFrame.HLine)
          sepline.setFrameShadow(QtWidgets.QFrame.Sunken)
          
          self.layout.addWidget(about, 0, 0)
          self.layout.addWidget(sepline, 1, 0, 1, 2)
          
          
          # loop over keys to add controls and validators
          nrow = len(keys)+1
          #qle - qlinedit dictionary to retrieve data later
          self.qle={}
          for i, key in enumerate(keys):
               Val_l  = QtWidgets.QLabel(labels[i], self)
               Val_t  = QtWidgets.QLineEdit(self)
               Val_t.setValidator(QtGui.QDoubleValidator(self))
               
               Val_t.textChanged.connect(self.check_state)
               Val_t.setText(data.get(key))
               self.layout.addWidget(Val_l, i+2, 0)
               self.layout.addWidget(Val_t, i+2, 1)
               
               self.qle[key]=Val_t
                       
          self.layout.addWidget(ok, nrow+2, 0)
          self.layout.addWidget(cancel, nrow+2, 1)
          
     def check_state(self, *args, **kwargs):
         
        sender = self.sender()
        validator = sender.validator()
        state = validator.validate(sender.text(), 0)[0]
        if state == QtGui.QValidator.Acceptable:
            color = '#c4df9b' # green
        elif state == QtGui.QValidator.Intermediate:
            color = '#fff79a' # yellow
        else:
            color = '#f6989d' # red
        sender.setStyleSheet('QLineEdit { background-color: %s }' % color)
        
     def OnOk(self):
         for i, key in enumerate(self.keys):
             try:
                 self.data[key]=self.qle[key].text()
             except:
                 mb=QtWidgets.QMessageBox(self)
                 mb.setWindowTitle('Entry error')
                 mb.setText("Please enter the missing parameter")
                 mb.exec_()
                 self.qle[key].setFocus()
                 return
         self.close()
     
     def OnCancel(self):
         self.destroy()




#----------------------------------------------------------------------
# main frame
#----------------------------------------------------------------------

class IAFrame(QtWidgets.QMainWindow):
    def __init__(self, parent, callback = None, **kwargs):
        super(IAFrame, self).__init__()
        # begin wxGlade: IAFrame.__init__
        self.setWindowTitle('Image Analyzer')
        self.setGeometry(150,150,700,600)
        
        # Menu Bar
        self.IAFrame_menubar = self.menuBar()
        self.createMenubar()
        
 # setup figure
        self.figure = Figure(figsize=(10,6))
        self.axes = self.figure.add_subplot(111)
        
        self.figure_canvas = FigureCanvas(self.figure)
        self.canvas =  self.figure_canvas
        
# setup callback function for MPL API event interface
        self.figure_canvas.mpl_connect('button_press_event', self.button_press_callback)
        self.figure_canvas.mpl_connect('button_release_event', self.button_release_callback)
        self.figure_canvas.mpl_connect('motion_notify_event', self.UpdateStatusBar)
        self.figure_canvas.mpl_connect('motion_notify_event', self.motion_notify_callback)
        self.figure_canvas.mpl_connect('figure_enter_event', self.ChangeCursor)
        self.setCentralWidget(self.figure_canvas)
        
# add ToolBar
        self.toolbar = NavigationToolbar2QT(self.figure_canvas, self, coordinates=False)
        self.addToolBar(self.toolbar)
        
# create status bar
        self.stBar1 = QtWidgets.QLabel('No file open')
        self.stBar1.setFrameStyle(2)
        self.stBar1.setFrameShadow(48)
        self.stBar2 = QtWidgets.QLabel('Pos:')
        self.stBar2.setFrameStyle(2)
        self.stBar2.setFrameShadow(48)
        self.stBar3 = QtWidgets.QLabel('Status: idle')
        self.stBar3.setFrameStyle(2)
        self.stBar3.setFrameShadow(48)
          
        self.statusBar().addWidget(self.stBar1, 1)
        self.statusBar().addWidget(self.stBar2, 1)
        self.statusBar().addWidget(self.stBar3, 1)
       
        self.image = None
        self.npoints = 0
        # array with the picked points
        self.points = []
        # array of calibration distances
        self.dx = []
        self.dy = []
        self.distance = []
        self.pix_distance = []
        self.cal_dx = []
        self.cal_dy = []
        # conversion pix to user (cal.) values
        self.convx = []
        self.convy = []
        # averaged conversion values
        self.convx_av = 1.
        self.convy_av = 1.
        self.enter_positions = False
        self.calibrate = False
        self.measure = False
        self.filename=""
        self.points_filename="points.data"
        
        self.button_release = False
        self.dlx = 0.02 #fractional offsets
        self.dly = 0.02
        self.dtx = 0.01
        self.dty = 0.01
#        # create a line for calibrated drawing later:
        self.lines = [] 
        self.leaders = []
        self.texts = []
        
        self.show()
        
    def closeEvent(self, event):
        close = QtWidgets.QMessageBox()
        close.setText("Are you sure?")
        close.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.Cancel)
        close = close.exec()
        if close == QtWidgets.QMessageBox.Yes:
            event.accept()
            QtWidgets.QApplication.quit()
        else:
            event.ignore()
        
    
    #menu bar data
    def menuData(self): # data for the menu
          return(
               ("File",                                        # menu label
                ("Open", "Open an image file to analyze", self.onOpen),  # menu items consisting of: label, text, handler
                ("Save", "Save positions", self.onSave),  
                ("Save as", "Save positions as", self.onSaveAs), 
                ("Quit", "Quit", self.onQuit), 
                ), # label, status and handler
               #
               ("&Analyze",
                ("&Pick points", "", self.onEnterPositions),
                ("&Measure", "", self.onMeasure),
#                ("&Redraw", "", self.onRedraw),
                ("&Clear", "", self.onClear),
                ),
               #
               ("&Calibration",
                ("&Calibrate", "Pick calibration points", self.onCalibrate), 
                ("&Load Calibration", "Get calibration information from file", self.onLoadCal), 
                ("&Enter Calibration", "Enter calibration parameters", self.onEnterCal), 
                ("&Show", "Display calibration parameters", self.onShowCal),
                ("&Clear", "Clear calibration", self.onClearCal)),
               )
               
    def createMenubar(self):
         for eachMenuData in self.menuData():
               
              menuLabel = eachMenuData[0]
              menuItems = eachMenuData[1:]
               

              menu = QtWidgets.QMenu(menuLabel, self)
              for eachItem in menuItems:
                   subMenu=QtWidgets.QAction(eachItem[0],self) 
                   subMenu.triggered.connect(eachItem[2])
                   menu.addAction(subMenu)
              
              self.IAFrame_menubar.addMenu(menu)
              


    def get_file_info(self, filename):
        dir, fname = os.path.split(filename)
        name, ext = os.path.splitext(fname)
        return dir, name, ext
        # that's it

    def update_graph(self):
        self.axes.clear()
        if not (self.image is None):
            self.axes.imshow(self.image, interpolation = None)
        self.figure.canvas.draw()
        # necessary to make a hard copy
        self.axes.set_autoscale_on(False)
        
    def onOpen(self):  # wxGlade: IAFrame.<event_handler>
        # get a filenmae
        filename = QtWidgets.QFileDialog.getOpenFileName(self, 'Select a file', os.getcwd())
        # read the image
        try:
            self.image = plt.imread(filename[0])
        except:
            if filename[0] == '':
                return
            else:
                print("cannot open: ", filename[0])
            
        self.filename = filename[0]
        dir, name, ext = self.get_file_info(self.filename)
        self.dir = dir+'//'
        self.name = name
        self.ext = ext
        # show current fil name
        self.stBar1.setText(self.name+self.ext)
        self.calibrate = False
        self.measure = False
        self.enter_positions = False
        self.update_status("idle")
        # show the image
        self.update_graph()

    def onLoadCal(self, event):  # wxGlade: IAFrame.<event_handler>
        # get the filename containing the calibration
        filename = QtWidgets.QFileDialog.getOpenFileName(self, 'Select calibration file', os.getcwd())
        # open file
        try:
            cal = B.get_file(filename[0])
#            xx = cal.par.get_value( 'convx_av')
        except:
            dlg = QtWidgets.QMessageBox(self)
            dlg.setText("Cannot find calibration information !")
            dlg.setWindowTitle("Calibration Problem")
            dlg.setIcon(2)
            dlg.exec_()
            return
        # get first calibration
        self.convx_av = cal.par.get_value( 'convx_av')
        self.convy_av = cal.par.get_value( 'convy_av')

    def onEnterCal(self):
        # enter calibration parameters
        # parameter keys
        pkeys=['conversion factor:']
        # current data to be shown in the dialog
        data = {pkeys[0]:"%f"%(self.convx_av), \
                }
        pdlg = NumberDialog(data, title="X-axis Calibration ", \
                            labels = pkeys, \
                            keys = pkeys, \
                            about_txt = "Enter the conversion factor")
        
        # now set the new parameters
        pdlg.exec_()
        self.convx_av = float(data[ pkeys[0] ])
        pdlg.destroy()
        # scale the y-axis
        # check if the same scale is to be used for y-axis
        dlg = QtWidgets.QMessageBox(self)
        dlg.setText("use the same as for X-axis ?")
        dlg.setWindowTitle("Y-axis scale")
        dlg.setStandardButtons(QtWidgets.QMessageBox.Yes|QtWidgets.QMessageBox.No)
        same_as_x = dlg.exec_() == QtWidgets.QMessageBox.Yes

        if same_as_x:
            self.convy_av = self.convx_av 
        else:
            # current data to be shown in the dialog
            data = {pkeys[0]:"%d"%(self.convy_av), \
                    }
            pdlg = NumberDialog(data, title="Y-axis Calibration ", \
                                labels = pkeys, \
                                keys = pkeys, \
                                about_txt = "Enter the conversion factor")
            
            pdlg.exec_()
            # now set the new parameters
            self.convy_av = float(data[ pkeys[0] ])
            pdlg.destroy()
        # done manually entering conversion factors
        
    def onShowCal(self, event):
        # show calibration parameters
        message = "convx_av = {0:10.3e}/ convy_av = {1:10.3e}".format(self.convx_av, self.convy_av)
        dlg = QtWidgets.QMessageBox(self)
        dlg.setText(message)
        dlg.setWindowTitle("Current conversion factors")
        dlg.exec_()

    def onClearCal(self):
        # clear the calibration point arrays and parameter arrays
        self.dx = []
        self.dy = []
        self.distance = []
        self.pix_distance = []
        self.cal_dx = []
        self.cal_dy = []
        # conversion pix to user (cal.) values
        self.convx = []
        self.convy = []
        # averaged conversion values
        self.convx_av = 1.
        self.convy_av = 1.

    def save_all(self):
        self.save_points(self.points_filename)
        dir, name, ext = self.get_file_info(self.points_filename)
        self.set_animated(False)
        self.figure.savefig(name + ".png")
        self.set_animated(True)

    def onSave(self):  # wxGlade: IAFrame.<event_handler>
        self.save_all()
        #event.Skip()


    def onSaveAs(self):  # wxGlade: IAFrame.<event_handler>
        file_dlg = QtWidgets.QFileDialog.getSaveFileName(self, "Save points as ...",
                            os.getcwd(),"*.data")
        if file_dlg[0] != '':
            filename = file_dlg[0]
            self.points_filename = filename
            self.save_all()

    def onQuit(self):  # wxGlade: IAFrame.<event_handler>
        self.close()
        
    def onCalibrate(self):  # wxGlade: IAFrame.<event_handler>
        if self.enter_positions:
            if self.warning( "You are still entering positions!", "stop and calibrate                            ?"):
                self.calibrate = True
                self.measure = False
                self.enter_positions = False
                self.update_status("calibrate")
            else:
                return
        self.calibrate = True
        self.measure = False
        self.enter_positions = False
        self.update_status("calibrate")
        
    def onEnterPositions(self):  # wxGlade: IAFrame.<event_handler>
        self.enter_positions = True
        self.measure = False
        self.calibrate = False
        self.update_status("enter positions")

    def onMeasure(self):
        if self.enter_positions:
            if self.warning( "You are still entering positions!","stop and measure                                ?"):
                self.measure = True
                self.enter_positions = False
                self.calibrate = False
                self.update_status("measure")
            else:
                return
        self.measure = True
        self.calibrate = False
        self.enter_positions = False
        self.update_status("measure")

#    def onRedraw(self):  # wxGlade: IAFrame.<event_handler>
#        self.redraw()

    def onClear(self):
        # reset markers
        self.npoints = 0
        # picked points
        self.points = []
        self.clear()


    def warning(self, title, message):
        dlg = QtWidgets.QMessageBox(self)
        dlg.setText(message)
        dlg.setWindowTitle(title)
        dlg.setStandardButtons(QtWidgets.QMessageBox.Ok|QtWidgets.QMessageBox.Cancel)
        ret = dlg.exec_() == QtWidgets.QMessageBox.Ok
        
        return ret

    def update_status(self, status):
        self.stBar3.setText("Status: " + status)

    def save_points(self, filename):
        f = open(filename,'w')
        # write header
        f.write("# data points from file : " +self.filename+'\n')
        f.write("# xpc and ypc are positions in calibrated user units (e.g mm) \n")
        f.write("#\n")
        f.write("# Calibration information \n")
        f.write("#\\ convx_av = {0:11.4e}\n".format(self.convx_av))
        f.write("#\\ convy_av = {0:11.4e}\n".format(self.convy_av))
        f.write("#! n[i,0]/ xp[f,1]/ yp[f,2]/ xpc[f,3]/ ypc[f,4]/ \n")
#        convx = 1.
#        convy = 1.
#        # use the first conversion factor by default
#        if self.convx != []:
#            convx = self.convx[0]
#            convy = self.convy[0]
        #
        for i,p in enumerate(self.points):
            f.write( "{0:d} {1:11.4e} {2:11.4e} {3:11.4e} {4:11.4e} \n".format( i, p[0], p[1], p[0]*self.convx_av, p[1]*self.convy_av ) )
        f.close()
        dir, name, ext = self.get_file_info(filename)
        self.update_status("saved to:  " + name+ext)        



###############################################################################
    
#from draw pannel class

###############################################################################

    def UpdateStatusBar(self, event):
        if event.inaxes:
            x, y = event.xdata, event.ydata
            text="Pos: {0:.2f}, {1:.2f}".format(x,y)
            self.stBar2.setText(text)
    
    def motion_notify_callback(self, event):
        if event.inaxes is None: return # do nothing if not in axis 
        if event.button != 1: return   # no nothing if the button is not pressed
        if self.calibrate and (not self.button_release):
            self.end_pos = event.xdata, event.ydata
            self.sc_end_pos = event.x, event.y
            self.canvas.restore_region(self.background)
            # data for get_line
            #x = [self.start_pos[0], self.end_pos[0]]
            #y = [self.start_pos[1], self.end_pos[1]]
            # data fot get_step_line
            x = [self.start_pos[0], self.end_pos[0], self.end_pos[0]]
            y = [self.start_pos[1], self.start_pos[1], self.end_pos[1]]
            self.current_line.set_data( (x,y))
            self.canvas.restore_region(self.background)
            self.axes.draw_artist(self.current_line)
            self.canvas.blit(self.axes.bbox)
        if self.measure and (not self.button_release):
            self.end_pos = event.xdata, event.ydata
            self.sc_end_pos = event.x, event.y
            self.canvas.restore_region(self.background)
            # data for get_line
            #x = [self.start_pos[0], self.end_pos[0]]
            #y = [self.start_pos[1], self.end_pos[1]]
            # data fot get_step_line
            x = [self.start_pos[0], self.end_pos[0]]
            y = [self.start_pos[1], self.end_pos[1]]
            self.current_line.set_data( (x,y))
            self.canvas.restore_region(self.background)
            self.axes.draw_artist(self.current_line)
            self.canvas.blit(self.axes.bbox)

    def ChangeCursor(self, event):
          self.figure_canvas.setCursor(QtGui.QCursor(QtCore.Qt.CrossCursor))
    
    def button_press_callback(self,event):
        # event.button : which button
        # event.x : screen x-position
        # event.y : screen y-position
        # event.xdata : x-position in axis units (None if outside)
        # event.ydata : y-position in axis units (None if outside)
        # event.inaxis: Axes in which the mouse currently is (None if outside)
        if event.inaxes is None:
            return
        if event.button > 1:
            if self.enter_positions:
                self.enter_positions = False
                self.update_status("idle")
                return
        pos = (event.xdata, event.ydata)
        self.background = self.canvas.copy_from_bbox(self.axes.bbox)
        if self.enter_positions:
            # enter positions
            # draw marker at the location
            self.npoints += 1
            self.points.append( pos )
            self.draw_marker(pos, self.npoints )
        elif self.calibrate:
            self.start_pos = event.xdata, event.ydata
            self.end_pos = event.xdata, event.ydata
            self.sc_start_pos = event.x, event.y
            self.sc_end_pos = event.x, event.y
            # self.current_line = self.get_line()
            self.current_line = self.get_step_line()
            self.lines.append(self.current_line)
        elif self.measure:
            self.start_pos = event.xdata, event.ydata
            self.end_pos = event.xdata, event.ydata
            self.sc_start_pos = event.x, event.y
            self.sc_end_pos = event.x, event.y
            # self.current_line = self.get_line()
            self.current_line = self.get_line()
            self.lines.append(self.current_line)
        # self.update_graph()
        
    def button_release_callback(self,event):
        # here we need a text control to enter the real distance for calibration
        # set this flag true to freeze the current mouse positions
        # necessary for  WINDOWS running as ShowModal does not completely work
        
        if self.calibrate:
            self.process_calibration()
        if self.measure:
            self.button_release = True
            self.process_measure()
            

    def set_animated(self, value):
        # change the status on animated, used value = False for hard copies
        for l in self.leaders:
            l.set_animated(value)
        for t in self.texts:
            t.set_animated(value)
        for L in self.lines:
            L.set_animated(value)
    
#    def redraw(self):
#        self.set_animated(False)
#        self.background = self.canvas.copy_from_bbox(self.axes.bbox)
#        self.set_animated(True)
#        self.canvas.restore_region(self.background)

    def clear(self):
        # reset all arrays
        self.lines = []
        self.texts = []
        self.leaders = []
        self.update_graph()            
    
    def get_line(self):
        # to draw animated lines
        x = [0.,1.]
        y = [0.,1.]
        line, = self.axes.plot(x,y, color = 'b', marker='x', animated=True) # do not show the line now !
        return line

    def get_step_line(self):
        # to draw amimated step line
        # three points
        x = [0.,1.,1.]
        y = [0.,0.,1.]
        line, = self.axes.plot(x,y, color = 'b', marker='x', animated=True) # do not show the line now !
        return line

    def get_text(self):
        text = self.axes.text(0,0, "", color = 'r', animated = True )
        return text
    
    def get_marker(self):
        # to draw animated markers
        x = [0.,1.]
        y = [0.,1.]
        leader, = self.axes.plot(x,y, color = 'r', animated = True)
        text = self.axes.text(0,0, "", color = 'r', animated = True )
        return (leader, text)
        


    def process_calibration(self):
        # get offsets in real world values
        start_pos = np.array(self.start_pos)
        end_pos = np.array(self.end_pos)
        sc_start_pos = np.array(self.sc_start_pos)
        sc_end_pos = np.array(self.sc_end_pos)
        r = end_pos - start_pos
        rs = sc_end_pos - sc_start_pos
        d = np.sqrt( np.sum(r*r) )
        dx = r[0]
        dy = r[1]
        self.distance.append(d)
        # same end point as starting point, give warning
        if d == 0.:
            self.warning( "Calibration Problem : Distance = 0.", \
                                 "You need to keep the mouse button pressed while dragging !")
            self.button_release = False
            return
        # use parameter
        self.dx.append(r[0])
        self.dy.append(r[1])
        ds = np.sqrt( np.sum(rs*rs) )
        self.pix_distance.append(ds)
        #
        # choice for calibrating
        choices = ["X-axis", "Y-axis", "X/Y independent"]
        selection = choices[0]
        dlg = QtWidgets.QInputDialog(self)
        dlg.setWindowTitle("Calibration Method " )
        dlg.setLabelText("Select axis to be used") 
        dlg.setComboBoxItems(choices)
        dlg.setOption(dlg.UseListViewForComboBoxItems)
        ok = dlg.exec_() == 1
        item = dlg.textValue()
        dlg.destroy()
        if ok and item!='':
            selection = item
        # scale the x-axis
        # use a NumberDialog to get calibration data
        # parameter keys
        pkeys=['Real (world units) displacement:']
        # current data to be shown in the dialog
        data = {pkeys[0]:"%d"%(r[0]), \
                }
        cdx = 1.
        cdy = 1.
        if (selection != choices[1]):
            pdlg = NumberDialog(data, title="X-axis Calibration ", \
                                labels = pkeys, \
                                keys = pkeys, \
                                about_txt = "Enter the measured displacement in world units")
            pdlg.exec_()
            pdlg.destroy()
            # now set the new parameters
            cdx = float(data[ pkeys[0] ])
        if (selection != choices[0]):
            # scale the y-axis
            # current data to be shown in the dialog
            data = {pkeys[0]:"%d"%(r[1]), \
                    }
            pdlg = NumberDialog(data, title="Y-axis Calibration ", \
                                labels = pkeys, \
                                keys = pkeys, \
                                about_txt = "Enter the measured displacement in world units")
            pdlg.exec_()
            pdlg.destroy()
            # now set the new parameters
            cdy = float(data[ pkeys[0] ])
        if selection == choices[0]:
            cdy = cdx/dx*dy
        if selection == choices[1]:
            cdx = cdy/dy*dx
        # update calibration factor
        self.convx.append( cdx/dx)
        self.convy.append( cdy/dy)
        # calcula averaged calibration
        self.convx_av = ( np.array(self.convx) ).mean()
        self.convy_av = ( np.array(self.convy) ).mean()
        # 
        self.cal_dx.append( cdx  )
        self.cal_dy.append( cdy  )
#        cd = np.sqrt(cdx**2 + cdy**2 )
        # create the text labels
        textx = self.get_text()
        texty = self.get_text()
        text_posx = start_pos + 0.5*np.array([dx, -0.1*dy])
        text_posy = end_pos - 0.5*np.array([0., dy])
        textx.set_position(text_posx )
        texty.set_position(text_posy )
        #text.set_rotation(text_angle)
        textx.set_ha('center')
        texty.set_ha('right')
        textx.set_rotation_mode('anchor')
        texty.set_rotation_mode('anchor')
        textx.set_text("dx {0:6.1f}/{1:10.3e}".format(r[0],cdx) )
        texty.set_text("dy {0:6.1f}/{1:10.3e}".format(r[1],cdy) )
        # draw the text
        self.axes.draw_artist(textx)
        self.axes.draw_artist(texty)
        self.texts.append(textx)
        self.texts.append(texty)
        self.canvas.blit(self.axes.bbox)
        # reset the calibration
        self.calibrate = False
        self.update_status("idle")
        # store last image
        self.background = self.canvas.copy_from_bbox(self.axes.bbox)
        self.button_release = False

    def process_measure(self):
        # get offsets in real world values
        start_pos = np.array(self.start_pos)
        end_pos = np.array(self.end_pos)
        sc_start_pos = np.array(self.sc_start_pos)
        sc_end_pos = np.array(self.sc_end_pos)
        r = end_pos - start_pos
        rs = sc_end_pos - sc_start_pos
        d = np.sqrt( np.sum(r*r) )
#        r_rot = np.array([-r[1], r[0]])
        dx = r[0]
        dy = r[1]
        cdx = dx * self.convx_av 
        cdy = dy * self.convy_av
        cd = np.sqrt(cdx**2 + cdy**2)
        self.distance.append(d)
        # same end point as starting point, give warning
        if d == 0.:
            self.warning( "Distance = 0.", \
                                 "You need to keep the mouse button pressed while dragging !")
            self.button_release = False
            return
        # use parameter
        self.dx.append(r[0])
        self.dy.append(r[1])
        ds = np.sqrt( np.sum(rs*rs) )
        self.pix_distance.append(ds)
        # create the text labels
        textm = self.get_text()
        text_posm = start_pos + 0.5*np.array([dx, dy]) + (0.03*np.array([-dy, -dx]))
        # calculate the text angle (use pixel coords.)
        text_angle = 0.
        if ds != 0.:
            txt_sin = rs[1]/ds
            txt_cos = rs[0]/ds
            text_angle = angle(txt_sin,txt_cos)
        textm.set_position(text_posm )
        textm.set_rotation(text_angle * 180./np.pi)
        textm.set_ha('center')
        textm.set_linespacing(2.)
        textm.set_rotation_mode('anchor')
        s_string = "\nD = {0:10.3e}/{1:6.1f}".format(cd,d)
        textm.set_text( s_string)
        # draw the text
        self.axes.draw_artist(textm)
        self.texts.append(textm)
        self.canvas.blit(self.axes.bbox)
        # reset the 
        self.measure = False
        self.update_status("idle")
        # store last image
        self.background = self.canvas.copy_from_bbox(self.axes.bbox)
        self.button_release = False
        

            
    def draw_marker(self, pos, n):
        # draw the n'th marker
        # draw a circle with a leader and the number n
        mx = pos[0]
        my = pos[1]
        # get offsets in real world values
        xmin,xmax = self.axes.get_xlim()
        ymin,ymax = self.axes.get_ylim()
        rx = xmax - xmin
        ry = ymax - ymin
        lpx = mx+self.dlx*rx
        lpy = my-self.dly*ry
        # the add text
        tpx = lpx + self.dtx*rx
        tpy = lpy - self.dty*ry
        text_string = "({0:d})".format(n)
        leader, text = self.get_marker()
        text.set_text(text_string)
        text.set_position( (tpx, tpy) )
        print("Marker:", mx, my, lpx, lpy, tpx, tpy)
        leader.set_data( [mx, lpx], [my, lpy])
        self.leaders.append(leader)
        self.texts.append(text)
        self.axes.draw_artist(leader)
        self.axes.draw_artist(text)
        self.canvas.blit(self.axes.bbox)


if __name__ == "__main__":
    app = QtCore.QCoreApplication.instance()
    if app is None:
        app = QtWidgets.QApplication(sys.argv)
    frame = IAFrame(app)
    print("ready for work")
    sys.exit(app.exec_())
